# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.exceptions import ensure_str


@pytest.mark.parametrize(
    "inp,exp",
    (("asd", "asd"), (b"asd", "asd"), ("🍺", "🍺"), (b"\xf0\x9f\x8d\xba", "🍺")),
    ids=("asd", "bytes-asd", "beer-str", "beer-bytes"),
)
def test_ensure_str(inp, exp):
    assert ensure_str(inp) == exp



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Generate initial dist-git configuration for packit's release syncing
"""

import logging
import os
from pathlib import Path
from typing import Any, Optional

import click
import yaml
from yaml import safe_load

from packit.api import PackitAPI
from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_existing_config
from packit.config import PackageConfig
from packit.config.config import Config, pass_config
from packit.distgit import DistGit
from packit.exceptions import PackitException
from packit.utils import commands

logger = logging.getLogger(__name__)

ONBOARD_BRANCH_NAME = "packit-config"

CONFIG_FILE_NAME = ".packit.yaml"
CONFIG_HEADER = """# See the documentation for more information:
# https://packit.dev/docs/configuration/
"""

COMMIT_MESSAGE = """Add Packit configuration for automating release syncing"""
PR_DESCRIPTION = """
For more details, see https://packit.dev/docs/configuration/ or contact
[the Packit team](https://packit.dev#contacts).

"""


@click.command("init", context_settings={"ignore_unknown_options": True})
@click.option(
    "--upstream-git-url",
    help="URL to the upstream GIT repository",
)
@click.option(
    "--upstream-git-url-command",
    help="Command to get the URL of the upstream git repository",
)
@click.option(
    "--upstream-tag-template",
    help="Template applied for upstream tags if they differ from versions. E.g. 'v{version}' ",
)
@click.option(
    "--upstream-tag-include",
    help="Python regex used for filtering upstream tags to include. ",
)
@click.option(
    "--upstream-tag-exclude",
    help="Python regex used for filtering upstream tags to exclude. ",
)
@click.option(
    "--version-update-mask",
    help="Python regex used for comparison of the old and the new version. ",
)
@click.option(
    "--issue-repository",
    help="URL of a git repository that can be used for reporting errors in form of issues. ",
)
@click.option(
    "--no-pull",
    default=False,
    is_flag=True,
    help="Do not include the pull from upstream job in the config",
)
@click.option(
    "--no-koji-build",
    default=False,
    is_flag=True,
    help="Do not include the Koji build job in the config",
)
@click.option(
    "--allowed-committers",
    help="Comma separated list of allowed_committers used for Koji builds",
    default="",
)
@click.option(
    "--allowed-pr-authors",
    help="Comma separated list of allowed_pr_authors used for Koji builds",
    default="",
)
@click.option(
    "--no-bodhi-update",
    default=False,
    is_flag=True,
    help="Do not include the Bodhi update job in the config",
)
@click.option(
    "--actions-file",
    help="Yaml file with 'actions' that should be used for the config",
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--dist-git-branches",
    help="Comma separated list of target branches in dist-git to release into. "
    "(defaults to rawhide)",
)
@click.option(
    "--push-to-distgit",
    "-p",
    default=False,
    is_flag=True,
    help="Push the generated Packit config to the dist-git repository's rawhide",
)
@click.option(
    "--create-pr",
    "-c",
    default=False,
    is_flag=True,
    help="Create a PR with generated Packit config",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Reset config to default if already exists.",
)
@click.option(
    "--clone-path",
    help="Path to clone the dist-git repo into (if path_or_url is URL). "
    "Otherwise clone the repo in a temporary directory.",
)
@click.option(
    "--commit-msg",
    help="Commit message used when creating a PR (also for the title) or pushing to dist-git. "
    f"Default: {COMMIT_MESSAGE!r}",
)
@click.argument("other_args", nargs=-1, type=click.UNPROCESSED)
@click.argument(
    "path_or_url",
    type=LocalProjectParameter(),
    default=os.path.curdir,
)
@pass_config
@cover_packit_exception
def init(
    config,
    upstream_git_url,
    upstream_git_url_command,
    upstream_tag_template,
    upstream_tag_include,
    upstream_tag_exclude,
    version_update_mask,
    issue_repository,
    no_pull,
    no_koji_build,
    allowed_committers,
    allowed_pr_authors,
    no_bodhi_update,
    actions_file,
    dist_git_branches,
    push_to_distgit,
    create_pr,
    force,
    clone_path,
    commit_msg,
    path_or_url,
    other_args,
):
    """
    Create the initial Packit dist-git configuration for Fedora release syncing based on
    the input parameters.

    This command adds `.packit.yaml` file to the dist-git repository either:

    \b 1. specified by path (defaults to current working directory)

    \b 2. specified by URL (`https://src.fedoraproject.org/rpms/<package>`) - clones the repository
    and adds the config in there. Ideally use this with --clone-path option, otherwise the
    repository is cloned to a temporary directory that is then removed.

    By default, all 3 jobs (`pull_from_upstream`, `koji_build`, `bodhi_update`) for release
    syncing are configured. You can use --no-pull, --no-koji-build or --no-bodhi-update
     options to not add some of the jobs (at the moment, Bodhi updates are created only for
     Koji builds built by Packit, therefore --no-koji-build implies also no Bodhi update job).

    You can either create the Packit config file only locally (default), or create a pull request
    (using --create-pr option) or push directly to the dist-git's default branch
    (--push-to-distgit).

    See 'packit init', if you want to initialize a repository as an upstream repo.

    Examples

    Local generation for dist-git repo in current working directory:

    \b
        $ packit dist-git init --upstream-git-url https://github.com/packit/packit .


    Local generation for dist-git repo specified by URL that will be cloned to `<my-package>` dir:

    \b
        $ packit dist-git init --upstream-git-url https://github.com/packit/packit --clone-path
        `<my-package>` https://src.fedoraproject.org/rpms/packit

    Using arbitrary configuration options that are not provided as the command options
    (the working dir needs to be specified in this case):

    \b
        $ packit dist-git init --upstream-git-url https://github.com/packit/packit
        --my-option option-value .

    """
    kwargs = {
        other_args[i][2:]: other_args[i + 1] for i in range(0, len(other_args), 2)
    }
    if no_pull and no_koji_build:
        logger.warning("At least one job needs to be defined!")
        return

    if upstream_git_url and upstream_git_url_command:
        click.echo(
            "Only one of --upstream-git-url or --upstream-git-url-command can be specified.",
            err=True,
        )
        return

    if upstream_git_url_command:
        click.echo(
            f"Running the following command to find upstream git URL: {upstream_git_url_command}",
        )
        result = commands.run_command(
            upstream_git_url_command,
            output=True,
            cwd=path_or_url.working_dir,
        )
        upstream_git_url = result.stdout.strip()
        click.echo(f"Found the following URL: {upstream_git_url}")

    DistGitInitializer(
        upstream_git_url=upstream_git_url,
        upstream_tag_template=upstream_tag_template,
        upstream_tag_include=upstream_tag_include,
        upstream_tag_exclude=upstream_tag_exclude,
        version_update_mask=version_update_mask,
        issue_repository=issue_repository,
        no_pull=no_pull,
        no_koji_build=no_koji_build,
        no_bodhi_update=no_bodhi_update,
        allowed_committers=allowed_committers,
        allowed_pr_authors=allowed_pr_authors,
        actions_file=actions_file,
        dist_git_branches=dist_git_branches,
        create_pr=create_pr,
        push_to_distgit=push_to_distgit,
        force=force,
        config=config,
        path_or_url=path_or_url,
        commit_msg=commit_msg,
        kwargs=kwargs,
    ).initialize_dist_git()


class PackitDumper(yaml.SafeDumper):
    # to not create yaml anchors when dumping
    # https://github.com/yaml/pyyaml/issues/535#issuecomment-1293636712
    def ignore_aliases(self, data):
        return True

    # correct list indentation
    # https://github.com/yaml/pyyaml/issues/234#issuecomment-765894586
    def increase_indent(self, flow=False, *args, **kwargs):
        return super().increase_indent(flow=flow, indentless=False)


class DistGitInitializer:
    def __init__(
        self,
        config: Config,
        path_or_url: LocalProjectParameter,
        upstream_git_url: Optional[str],
        upstream_tag_template: Optional[str] = None,
        upstream_tag_include: Optional[str] = None,
        upstream_tag_exclude: Optional[str] = None,
        version_update_mask: Optional[str] = None,
        issue_repository: Optional[str] = None,
        no_pull: bool = False,
        no_koji_build: bool = False,
        allowed_committers: Optional[str] = None,
        allowed_pr_authors: Optional[str] = None,
        no_bodhi_update: bool = False,
        actions_file: Optional[Path] = None,
        dist_git_branches: Optional[str] = None,
        push_to_distgit: bool = False,
        create_pr: bool = False,
        force: bool = False,
        commit_msg: Optional[str] = None,
        kwargs: Optional[dict] = None,
    ):
        self.config = config
        self.upstream_git_url = upstream_git_url
        self.upstream_tag_template = upstream_tag_template
        self.upstream_tag_include = upstream_tag_include
        self.upstream_tag_exclude = upstream_tag_exclude
        self.version_update_mask = version_update_mask
        self.issue_repository = issue_repository
        self.no_pull = no_pull
        self.no_koji_build = no_koji_build
        self.allowed_committers = (
            allowed_committers.split(",") if allowed_committers else None
        )
        self.allowed_pr_authors = (
            allowed_pr_authors.split(",") if allowed_pr_authors else None
        )
        self.no_bodhi_update = no_bodhi_update
        self.actions_file = actions_file
        self.dist_git_branches = (
            dist_git_branches.split(",") if dist_git_branches else ["fedora-rawhide"]
        )
        self.push_to_distgit = push_to_distgit
        self.create_pr = create_pr
        self.path_or_url = path_or_url
        self.force = force
        self.commit_msg = commit_msg or COMMIT_MESSAGE
        self.kwargs = kwargs or {}

    @property
    def working_dir(self):
        return self.path_or_url.working_dir

    @property
    def config_path(self):
        return self.get_or_create_config_path()

    @property
    def package_config_dict(self):
        return self.generate_package_config_dict()

    @property
    def actions(self):
        return self.parse_actions_from_file() if self.actions_file else {}

    @property
    def package_config_content(self):
        return (
            f"{CONFIG_HEADER}\n"
            f"{yaml.dump(self.package_config_dict, sort_keys=False, Dumper=PackitDumper)}"
        )

    def initialize_dist_git(self):
        logger.info(
            f"Generating config for dist-git repository placed in {self.config_path}",
        )

        logger.info(f"Generated config: \n\n{self.package_config_content}\n\n")

        if not (self.push_to_distgit or self.create_pr):
            self.write_package_config()
            return

        logger.info(
            f"About to {'push' if self.push_to_distgit else 'create PR with'} "
            f"the generated Packit config.",
        )

        self.write_and_push()

    def get_or_create_config_path(self):
        config_path = get_existing_config(self.working_dir)
        if config_path:
            if not self.force:
                raise PackitException(
                    f"Packit config {config_path} already exists."
                    " If you want to regenerate it use `--force` option",
                )
        else:
            config_path = self.working_dir / CONFIG_FILE_NAME

        return config_path

    def parse_actions_from_file(self) -> dict:
        with open(self.actions_file) as file:
            actions_content = file.read()

        try:
            actions = safe_load(actions_content)
            if not isinstance(actions, dict):
                raise ValueError("The content of the actions file is not a dictionary.")
        except Exception as e:
            raise ValueError(f"Error parsing YAML content: {e}") from e

        return actions

    def generate_package_config_dict(self):
        config: dict[str, Any] = {}

        options = [
            "upstream_git_url",
            "upstream_tag_template",
            "upstream_tag_include",
            "upstream_tag_exclude",
            "version_update_mask",
            "issue_repository",
            "allowed_committers",
            "allowed_pr_authors",
            "actions",
        ]

        config_key_mappings = {"upstream_git_url": "upstream_project_url"}

        for key in options:
            value = getattr(self, key, None)
            if value:
                config_key = (
                    config_key_mappings[key] if key in config_key_mappings else key
                )
                config[config_key] = value

        config.update(self.kwargs)

        config["jobs"] = []
        if not self.no_pull:
            config["jobs"].append(
                {
                    "job": "pull_from_upstream",
                    "trigger": "release",
                    "dist_git_branches": self.dist_git_branches,
                },
            )
        if not self.no_koji_build:
            config["jobs"].append(
                {
                    "job": "koji_build",
                    "trigger": "commit",
                    "dist_git_branches": self.dist_git_branches,
                },
            )

        # currently it doesn't make sense to create Bodhi job
        # if Koji job is not configured
        if not self.no_koji_build and not self.no_bodhi_update:
            config["jobs"].append(
                {
                    "job": "bodhi_update",
                    "trigger": "commit",
                    # TODO we could compute the branches to exclude from Bodhi (autoupdates)
                    "dist_git_branches": self.dist_git_branches,
                },
            )

        return config

    def write_and_push(self):
        # needed for PackitAPI to work
        package_config_dict_for_api = self.package_config_dict
        package_config_dict_for_api[
            "specfile_path"
        ] = f"{self.path_or_url.repo_name}.spec"

        package_config = PackageConfig.get_from_dict(
            raw_dict=package_config_dict_for_api,
            repo_name=self.path_or_url.repo_name,
        )

        api = PackitAPI(
            config=self.config,
            package_config=package_config,
            upstream_local_project=None,
            downstream_local_project=self.path_or_url,
        )

        default_dg_branch = api.dg.local_project.git_project.default_branch

        logger.info(f"Updating the default branch {default_dg_branch!r} first.")
        api.dg.update_branch(default_dg_branch)
        api.dg.switch_branch(default_dg_branch)

        if self.create_pr:
            self.handle_pr_creation(api, default_dg_branch)
        else:
            self.handle_push(api, default_dg_branch)

    def handle_pr_creation(self, api: PackitAPI, default_dg_branch: str):
        api.dg.create_branch(ONBOARD_BRANCH_NAME)
        api.dg.switch_branch(ONBOARD_BRANCH_NAME)
        self.write_package_config()
        self.commit_config(api.dg)
        api.push_and_create_pr(
            pr_title=self.commit_msg,
            pr_description=PR_DESCRIPTION,
            repo=api.dg,
            git_branch=default_dg_branch,
        )

    def handle_push(self, api: PackitAPI, default_dg_branch: str):
        self.write_package_config()
        self.commit_config(api.dg)
        api.dg.push(refspec=f"HEAD:{default_dg_branch}")

    def write_package_config(self):
        logger.info(f"Writing config to {self.config_path}")
        self.config_path.write_text(self.package_config_content)

    def commit_config(self, dist_git: DistGit):
        logger.info("About to add and commit the config.")
        dist_git.local_project.git_repo.index.add([CONFIG_FILE_NAME])
        dist_git.local_project.git_repo.index.commit(self.commit_msg)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.api import PackitAPI, SynchronizationStatus
from packit.constants import DISTRO_DIR, FROM_DIST_GIT_TOKEN, FROM_SOURCE_GIT_TOKEN
from packit.exceptions import PackitException
from tests.integration.conftest import mock_spec_download_remote_s


@pytest.fixture
def check_ready_api_dg_first(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_update_source_git,
) -> PackitAPI:
    """Fixture returning an API instance where dist-git and source-git
    repos contain commits with git-trailers. The latest trailing commit
    is source-git update (From-dist-git trailer)."""
    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)
    api_instance_update_source_git.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        upstream_ref="0.1.0",
        mark_commit_origin=True,
    )
    (distgit / "file").write_text("foo")
    api_instance_update_source_git.dg.commit("add new file", "")
    api_instance_update_source_git.update_source_git("HEAD~..")
    return api_instance_update_source_git


@pytest.fixture
def check_ready_api_sg_first(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_update_source_git,
) -> PackitAPI:
    """Fixture returning an API instance where dist-git and source-git
    repos contain commits with git-trailers. The latest trailing commit
    is dist-git update (From-source-git trailer)."""
    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    (distgit / "file").write_text("foo")
    api_instance_update_source_git.dg.commit("add new file", "")
    api_instance_update_source_git.update_source_git("HEAD~..")
    mock_spec_download_remote_s(distgit)
    api_instance_update_source_git.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        upstream_ref="0.1.0",
        mark_commit_origin=True,
    )
    return api_instance_update_source_git


def test_source_git_status_no_trailers(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_source_git,
):
    """Check that an error is thrown if no trailers are present."""
    with pytest.raises(PackitException):
        api_instance_source_git.sync_status()


def test_source_git_status_dg_commit_not_exists(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_source_git,
):
    """Check that an error is thrown if dist-git commit referenced
    by a From-dist-git trailer does not exist."""
    sourcegit, _ = sourcegit_and_remote
    (sourcegit / "file").write_text("foo")
    api_instance_source_git.up.commit(
        "change",
        "",
        trailers=[(FROM_SOURCE_GIT_TOKEN, "abcd")],
    )
    with pytest.raises(PackitException):
        api_instance_source_git.sync_status()


def test_source_git_status_sg_commit_not_exists(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_source_git,
):
    """Check that an error is thrown if source-git commit referenced
    by a From-source-git trailer does not exist."""
    distgit, _ = distgit_and_remote
    (distgit / "file").write_text("foo")
    api_instance_source_git.dg.commit(
        "change",
        "",
        trailers=[(FROM_DIST_GIT_TOKEN, "abcd")],
    )
    with pytest.raises(PackitException):
        api_instance_source_git.sync_status()


@pytest.mark.parametrize(
    "api",
    ["check_ready_api_dg_first", "check_ready_api_sg_first"],
)
def test_source_git_status_synced(
    sourcegit_and_remote,
    distgit_and_remote,
    api,
    request,
):
    """Dist-git and source-git are in sync."""
    api_instance = request.getfixturevalue(api)
    assert api_instance.sync_status() == SynchronizationStatus(None, None)


@pytest.mark.parametrize(
    "api",
    ["check_ready_api_dg_first", "check_ready_api_sg_first"],
)
def test_source_git_status_dist_git_ahead(
    sourcegit_and_remote,
    distgit_and_remote,
    api,
    request,
):
    """Dist-git has extra commits that must be synced."""
    api_instance = request.getfixturevalue(api)
    distgit, _ = distgit_and_remote
    (distgit / "file").write_text("aaa")
    api_instance.dg.commit("changes", "")
    range_start = api_instance.dg.local_project.git_repo.head.commit.hexsha
    (distgit / "file").write_text("aaaa")
    api_instance.dg.commit("changes", "")
    assert api_instance.sync_status() == SynchronizationStatus(None, range_start)


@pytest.mark.parametrize(
    "api",
    ["check_ready_api_dg_first", "check_ready_api_sg_first"],
)
def test_source_git_status_source_git_ahead(
    sourcegit_and_remote,
    distgit_and_remote,
    api,
    request,
):
    """Source-git has extra commits that must be synced."""
    api_instance = request.getfixturevalue(api)
    sourcegit, _ = sourcegit_and_remote
    (sourcegit / DISTRO_DIR / "new_file").write_text("bbbb")
    api_instance.up.commit("changes to source-git", "")
    sg_range_start = api_instance.up.local_project.git_repo.head.commit.hexsha
    assert api_instance.sync_status() == SynchronizationStatus(sg_range_start, None)


@pytest.mark.parametrize(
    "api",
    ["check_ready_api_dg_first", "check_ready_api_sg_first"],
)
def test_source_git_status_history_diverges(
    sourcegit_and_remote,
    distgit_and_remote,
    api,
    request,
):
    """Both source-git and dist-git have extra commits that must be
    synced (diversion)."""
    api_instance = request.getfixturevalue(api)
    distgit, _ = distgit_and_remote
    sourcegit, _ = sourcegit_and_remote
    (distgit / "file").write_text("aaa")
    api_instance.dg.commit("changes", "")
    dg_range_start = api_instance.dg.local_project.git_repo.head.commit.hexsha
    (distgit / "file").write_text("aaaa")
    api_instance.dg.commit("changes", "")

    (sourcegit / DISTRO_DIR / "new_file").write_text("bbbb")
    api_instance.up.commit("changes to source-git", "")
    sg_range_start = api_instance.up.local_project.git_repo.head.commit.hexsha
    assert api_instance.sync_status() == SynchronizationStatus(
        sg_range_start,
        dg_range_start,
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import re
import tempfile
from collections.abc import Iterable, Sequence
from functools import partial
from pathlib import Path
from typing import Optional, Union

import cccolutils
import git
from bodhi.client.bindings import BodhiClientException
from lazy_object_proxy import Proxy
from ogr.abstract import PullRequest
from ogr.services.pagure import PagureProject
from specfile.utils import NEVR

from packit.base_git import PackitRepositoryBase
from packit.command_handler import CommandHandler
from packit.config import (
    Config,
    MultiplePackages,
    PackageConfig,
    get_local_package_config,
)
from packit.constants import EXISTING_BODHI_UPDATE_REGEX
from packit.exceptions import (
    PackitBodhiException,
    PackitConfigException,
    PackitException,
)
from packit.local_project import CALCULATE, LocalProject, LocalProjectBuilder
from packit.pkgtool import PkgTool
from packit.utils.bodhi import get_bodhi_client
from packit.utils.commands import cwd
from packit.utils.koji_helper import KojiHelper
from packit.utils.lookaside import LookasideCache

logger = logging.getLogger(__name__)


class DistGit(PackitRepositoryBase):
    """
    A class which interacts with dist-git and pagure-over-dist-git API.

    The logic covers git and pagure interaction, manipulation with content,
    spec files, patches and archives.

    The expectation is that this class interacts with low level APIs (ogr) and
    doesn't hold any workflow; the workflow and feeding input should be done up the stack.

    The class works with a single instance of dist-git, that's the state,
    and methods of this class interact with the local copy.
    """

    # spec files are stored in this dir in dist-git
    # this applies to Fedora and CentOS Stream 9
    spec_dir_name = ""

    # sources are stored in this dir in dist-git
    # this applies to Fedora and CentOS Stream 9
    source_dir_name = ""

    def __init__(
        self,
        config: Config,
        package_config: MultiplePackages,
        local_project: LocalProject = None,
        clone_path: Optional[str] = None,
    ):
        """
        Args:
            config: User configuration of Packit.
            package_config: Packit configuration of the package
                related to this dist-git repo.
            local_project: LocalProject object.
            clone_path: Path where the dist-git repository is cloned.
        """
        super().__init__(config=config, package_config=package_config)

        self._local_project = local_project

        self.fas_user = self.config.fas_user
        self._downstream_config: Optional[PackageConfig] = None
        self._clone_path: Optional[str] = clone_path

    def __repr__(self):
        return (
            "DistGit("
            f"config='{self.config}', "
            f"package_config='{self.package_config}', "
            f"local_project='{self.local_project}', "
            f"downstream_config='{self.downstream_config}', "
            f"absolute_specfile_path='{self.absolute_specfile_path}')"
        )

    @classmethod
    def clone(
        cls,
        config: Config,
        package_config: MultiplePackages,
        path: Path,
        branch: Optional[str] = None,
    ) -> "DistGit":
        """
        Clone dist-git repo for selected package and return this class
        Args:
            config: global packit config
            package_config: package config: downstream_package_name is utilized for cloning
            path: clone the repo to this path
            branch: optionally, check out this branch
        Returns: instance of the DistGit class
        """
        dg = cls(config, package_config)
        dg.clone_package(target_path=path, branch=branch)
        dg._local_project = LocalProjectBuilder().build(
            working_dir=path,
            git_repo=CALCULATE,
        )
        return dg

    @property
    def local_project(self):
        """return an instance of LocalProject"""
        if self._local_project is None:
            if self._clone_path:
                working_dir = self._clone_path
            else:
                tmpdir = tempfile.mkdtemp(prefix="packit-dist-git")
                self.clone_package(target_path=tmpdir)
                working_dir = tmpdir

            self._local_project = LocalProject(
                working_dir=working_dir,
                git_url=self.package_config.dist_git_package_url,
                namespace=self.package_config.dist_git_namespace,
                repo_name=self.package_config.downstream_package_name,
                git_project=self.config.get_project(
                    url=self.package_config.dist_git_package_url,
                ),
                cache=self.repository_cache,
            )
            self._local_project.working_dir_temporary = not self._clone_path
            self._local_project.refresh_the_arguments()
            # TODO: Turn this on once p-s mocks are updated
            # builder = LocalProjectBuilder(cache=self.repository_cache)
            # self._local_project = builder.build(
            #    working_dir=working_dir,
            #    git_url=self.package_config.dist_git_package_url,
            #    namespace=self.package_config.dist_git_namespace,
            #    repo_name=self.package_config.downstream_package_name,
            #    git_project=self.config.get_project(
            #        url=self.package_config.dist_git_package_url
            #    ),
            #    git_repo=CALCULATE,
            # )
        elif not self._local_project.git_project:
            self._local_project.git_project = self.config.get_project(
                url=self.package_config.dist_git_package_url,
            )

        return self._local_project

    @property
    def command_handler(self) -> CommandHandler:
        if self._command_handler is None:
            self._command_handler = self.handler_kls(
                # so that the local_project is evaluated only when needed
                local_project=Proxy(partial(DistGit.local_project.__get__, self)),  # type: ignore
                config=self.config,
            )
        return self._command_handler

    @property
    def downstream_config(self) -> Optional[PackageConfig]:
        if not self._downstream_config:
            try:
                self._downstream_config = get_local_package_config(
                    self.local_project.working_dir,
                    repo_name=self.local_project.repo_name,
                )
            except PackitConfigException:
                return None
        return self._downstream_config

    @property
    def pkg_tool(self) -> str:
        """Returns the packaging tool. Prefers the package-level override."""
        return self.package_config.pkg_tool or self.config.pkg_tool

    def clone_package(
        self,
        target_path: Union[Path, str],
        branch: Optional[str] = None,
    ) -> None:
        """
        Clone package from dist-git, i.e. from:
        - Fedora's src.[stg.]fedoraproject.org
        - CentOS Stream's gitlab.com/redhat/centos-stream/rpms/
        - CentOS CBS's gitlab.com/CentOS/<SIG>/rpms/
        depending on configured pkg_tool: {fedpkg(default),fedpkg-stage,centpkg,cbs}

        Args:
            target_path: the name of a new directory to clone into
            branch: optional, branch to checkout
        """
        pkg_tool = PkgTool(
            fas_username=self.fas_user,
            directory=target_path,
            tool=self.pkg_tool,
        )
        pkg_tool.clone(
            package_name=self.package_config.downstream_package_name,
            target_path=target_path,
            branch=branch,
            anonymous=not cccolutils.has_creds(),
        )

    def get_absolute_specfile_path(self) -> Path:
        """provide the path, don't check it"""
        if not self.package_config.downstream_package_name:
            raise PackitException(
                "Unable to find specfile in dist-git: "
                "please set downstream_package_name in your .packit.yaml",
            )
        return (
            self.local_project.working_dir
            / self.spec_dir_name
            / f"{self.package_config.downstream_package_name}.spec"
        )

    @property
    def absolute_source_dir(self) -> Path:
        """absolute path to directory with spec-file sources"""
        return self.local_project.working_dir / self.source_dir_name

    @property
    def absolute_specfile_path(self) -> Path:
        if not self._specfile_path:
            self._specfile_path = self.get_absolute_specfile_path()
            if not self._specfile_path.exists():
                raise FileNotFoundError(f"Specfile {self._specfile_path} not found.")
        return self._specfile_path

    def update_branch(self, branch_name: str):
        """
        Fetch latest commits to the selected branch; tracking needs to be set up

        :param branch_name: name of the branch to check out and fetch
        """
        logger.debug(f"About to update branch {branch_name!r}.")
        origin = self.local_project.git_repo.remote("origin")
        origin.fetch()
        try:
            head = self.local_project.git_repo.heads[branch_name]
        except IndexError as e:
            raise PackitException(f"Branch {branch_name!r} does not exist.") from e
        try:
            remote_ref = origin.refs[branch_name]
        except IndexError as e:
            raise PackitException(
                f"Branch {branch_name} does not exist in the origin remote.",
            ) from e
        head.set_commit(remote_ref)

    def push_to_fork(
        self,
        branch_name: str,
        fork_remote_name: str = "fork",
        force: bool = False,
        sync_acls: bool = False,
    ):
        """
        push changes to a fork of the dist-git repo; they need to be committed!

        Args:
            branch_name: the branch where we push
            fork_remote_name: local name of the remote where we push to
            force: push forcefully?
            sync_acls: whether to sync the ACLs of the original repo and the fork

        """
        logger.debug(
            f"About to {'force ' if force else ''}push changes to branch {branch_name!r} "
            f"of a fork {fork_remote_name!r} of the dist-git repo.",
        )
        if fork_remote_name not in [
            remote.name for remote in self.local_project.git_repo.remotes
        ]:
            fork = self.local_project.git_project.get_fork()
            if not fork:
                self.local_project.git_project.fork_create()
                fork = self.local_project.git_project.get_fork()
            if not fork:
                raise PackitException(
                    "Unable to create a fork of repository "
                    f"{self.local_project.git_project.full_repo_name}",
                )
            if sync_acls and isinstance(self.local_project.git_project, PagureProject):
                # synchronize ACLs between original repo and fork for Pagure
                self.sync_acls(self.local_project.git_project, fork)
            fork_urls = fork.get_git_urls()
            self.local_project.git_repo.create_remote(
                name=fork_remote_name,
                url=fork_urls["ssh"],
            )

        try:
            self.push(refspec=branch_name, remote_name=fork_remote_name, force=force)
        except git.GitError as ex:
            msg = (
                f"Unable to push to remote fork {fork_remote_name!r} using branch {branch_name!r}, "
                f"the error is:\n{ex}"
            )
            raise PackitException(msg) from ex

    def create_pull(
        self,
        pr_title: str,
        pr_description: str,
        source_branch: str,
        target_branch: str,
    ) -> PullRequest:
        """
        Create dist-git pull request using the requested branches
        """
        logger.debug(
            "About to create dist-git pull request "
            f"from {source_branch!r} to {target_branch!r}.",
        )
        project = self.local_project.git_project

        project_fork = project.get_fork()
        if not project_fork:
            project.fork_create()
            project_fork = project.get_fork()

        try:
            dist_git_pr = project_fork.create_pr(
                title=pr_title,
                body=pr_description,
                source_branch=source_branch,
                target_branch=target_branch,
            )
        except Exception as ex:
            logger.error(f"There was an error while creating the PR: {ex!r}")
            raise

        logger.info(f"PR created: {dist_git_pr.url}")
        return dist_git_pr

    @property
    def upstream_archive_names(self) -> list[str]:
        """
        Sources that are considered upstream and should be uploaded to lookaside.
        Currently that's the source identified by spec_source_id (Source0 by default)
        and all other sources specified as URLs in the spec file.

        Returns:
             names of the archives, e.g. ['sen-0.6.1.tar.gz']
        """
        with self.specfile.sources() as sources:
            archive_names = [
                s.expanded_filename
                for s in sources
                if (s.remote or s.number == self.package_config.spec_source_id_number)
                and s.valid
            ]
        logger.debug(
            f"Remote sources or sources specified by spec_source_id (Source0 by default): "
            f"{archive_names}",
        )
        return archive_names

    @property
    def local_archive_names(self) -> list[str]:
        """
        Sources that are considered local (are not specified by URL).
        Source specified by spec_source_id is excluded as it is
        handled as upstream archive in upstream_archive_names.

        Returns:
             names of the archives, e.g. ['sen-0.6.1.tar.gz']
        """
        with self.specfile.sources() as sources:
            archive_names = [
                s.expanded_filename
                for s in sources
                if (
                    not s.remote
                    and s.number != self.package_config.spec_source_id_number
                )
                and s.valid
            ]
        logger.debug(
            f"Local sources (source specified by spec_source_id excluded): {archive_names}",
        )
        return archive_names

    def download_upstream_archives(self) -> list[Path]:
        """
        Fetch archives for the current upstream release defined in dist-git's spec

        :return: list of path to the archives
        """
        archives = []
        logger.info(f"Downloading archives: {self.upstream_archive_names}")
        with cwd(self.local_project.working_dir):
            self.download_remote_sources(self.pkg_tool)
        for upstream_archive_name in self.upstream_archive_names:
            archive = self.absolute_source_dir / upstream_archive_name
            if not archive.exists():
                raise PackitException(
                    "Upstream archive was not downloaded. Check that {} exists in dist-git.".format(
                        upstream_archive_name,
                    ),
                )
            archives.append(archive)
        return archives

    def download_source_files(self, pkg_tool: str = ""):
        """Download source files from the lookaside cache

        Use the pkg_tool that was specified.

        Args:
            pkg_tool: Name of the executable to be used to fetch
                the sources.
        """
        logger.info("Downloading source files from the lookaside cache...")
        pkg_tool_ = PkgTool(
            fas_username=self.config.fas_user,
            directory=self.local_project.working_dir,
            tool=pkg_tool or self.pkg_tool,
        )
        pkg_tool_.sources()

    def upload_to_lookaside_cache(
        self,
        archives: Iterable[Path],
        pkg_tool: str = "",
        offline: bool = False,
    ) -> None:
        """Upload files (archives) to the lookaside cache.

        If the archive is already uploaded, the rpkg tool doesn't do anything.

        Args:
            archive: Path to archive to upload to lookaside cache.
            pkg_tool: Optional, rpkg tool (fedpkg/centpkg/cbs) to use to upload.
            offline: Whether to use offline mode of the tool
                     (no actual upload, just local file updates).

        Raises:
            PackitException, if the upload fails.
        """
        logger.info("About to upload to lookaside cache.")
        logger.debug(f"Archives to upload: {archives}")
        pkg_tool_ = PkgTool(
            fas_username=self.config.fas_user,
            directory=self.local_project.working_dir,
            tool=pkg_tool or self.pkg_tool,
        )
        try:
            pkg_tool_.new_sources(
                sources=archives,
                offline=offline,
            )
        except Exception as ex:
            logger.error(
                f"'{pkg_tool_.tool} new-sources' failed for the following reason: {ex!r}",
            )
            raise PackitException(ex) from ex

    def is_archive_in_lookaside_cache(self, archive_path: Union[Path, str]) -> bool:
        """
        Check whether the archive is already uploaded to the lookaside cache.

        Args:
            archive_path: Path to the archive.

        Returns:
            `True`, if archive is present in the lookaside cache, `False`
            otherwise.
        """
        return LookasideCache(self.pkg_tool).is_archive_uploaded(
            self.package_config.downstream_package_name,
            archive_path,
        )

    def purge_unused_git_branches(self):
        # TODO: remove branches from merged PRs
        raise NotImplementedError("not implemented yet")

    def build(
        self,
        scratch: bool = False,
        nowait: bool = False,
        koji_target: Optional[str] = None,
    ):
        """
        Perform a `fedpkg build` in the repository

        :param scratch: should the build be a scratch build?
        :param nowait: don't wait on build?
        :param koji_target: koji target to pick (see `koji list-targets`)

        Returns:
            The 'stdout' of the build command.
        """
        pkg_tool = PkgTool(
            fas_username=self.fas_user,
            directory=self.local_project.working_dir,
            tool=self.pkg_tool,
        )
        return pkg_tool.build(scratch=scratch, nowait=nowait, koji_target=koji_target)

    @staticmethod
    def get_latest_build_for_branch(downstream_package_name, dist_git_branch):
        """Queries Koji for the latest build of a package for a dist-git branch.

        Args:
            downstream_package_name: Downstream package name.
            dist_git_branch: Associated dist-git branch.

        Returns:
            NVR of the latest build found.

        Raises:
            PackitException if there is no such build.
        """
        logger.debug(
            "Querying Koji for the latest build "
            f"of package {downstream_package_name!r} "
            f"for dist-git-branch {dist_git_branch!r}",
        )

        koji_helper = KojiHelper()
        if not (tag := koji_helper.get_candidate_tag(dist_git_branch)):
            raise PackitException(f"Failed to get candidate tag for {dist_git_branch}")
        build = koji_helper.get_latest_nvr_in_tag(downstream_package_name, tag)

        if not build:
            raise PackitException(
                f"There is no build for {downstream_package_name!r} "
                f"and koji tag {tag}",
            )
        logger.info(
            "Koji build for package "
            f"{downstream_package_name!r} and koji tag {tag}:"
            f"\n{build}",
        )
        return build

    @staticmethod
    def get_changelog_since_latest_stable_build(
        package: str,
        nvr: str,
    ) -> Optional[str]:
        """
        Retrieves changelog diff between the latest stable (tagged for a release)
        build and a build with the specified NVR.

        Args:
            package: Downstream package name.
            nvr: NVR of a build for which to get changelog diff.

        Returns:
            Changelog diff as a string or None if it wasn't possible to retrieve it.
        """
        koji_helper = KojiHelper()
        stable_tags = []
        for tag in koji_helper.get_build_tags(nvr):
            stable_tags = koji_helper.get_stable_tags(tag)
            if stable_tags:
                break
        latest_stable_nvr = None
        for tag in stable_tags:
            build = koji_helper.get_latest_nvr_in_tag(package, tag)
            if build:
                latest_stable_nvr = build
                break
        if not latest_stable_nvr or latest_stable_nvr == nvr:
            return None
        changelog = koji_helper.get_build_changelog(latest_stable_nvr)
        if not changelog:
            return None
        since = changelog[0][0]
        changelog = koji_helper.get_build_changelog(nvr)
        return koji_helper.format_changelog(changelog, since)

    def create_bodhi_update(
        self,
        dist_git_branch: str,
        update_type: str,
        update_notes: Optional[str] = None,
        koji_builds: Optional[Sequence[str]] = None,
        sidetag: Optional[str] = None,
        bugzilla_ids: Optional[list[int]] = None,
        alias: Optional[str] = None,
    ) -> Optional[tuple[str, str]]:
        """
        Create bodhi update.

        Args:
            dist_git_branch: Git reference.
            update_type: Type of the update, check CLI.
            update_notes: Notes about the update to be displayed in Bodhi. If not specified,
              automatic update notes including a changelog diff since the latest stable build
              will be generated.
            koji_builds: List of Koji builds or `None` (picks latest).
            sidetag: Koji sidetag to create the update from.
            bugzilla_ids: List of Bugzillas that are resolved with the update.
            alias: Alias of an existing update to edit. If not specified,
              a new update will be created.

        Returns:
            Alias and URL of the update or None if the update was already created.
        """
        logger.debug(
            f"About to create a Bodhi update of type {update_type!r} from {dist_git_branch!r}",
        )

        bodhi_client = get_bodhi_client()
        # make sure we have the credentials
        bodhi_client.ensure_auth()

        if not koji_builds:
            koji_builds = [
                self.get_latest_build_for_branch(
                    self.package_config.downstream_package_name,
                    dist_git_branch=dist_git_branch,
                ),
            ]

        bugzilla_ids_from_changelog = []

        # I was thinking of verifying that the build is valid for a new bodhi update
        # but in the end it's likely a waste of resources since bodhi will tell us
        if update_notes is not None:
            rendered_note = update_notes.format(version=self.specfile.expanded_version)
        else:
            builds = ", ".join(koji_builds)
            rendered_note = f"Automatic update for {builds}."
            for nvr in koji_builds:
                package = NEVR.from_string(nvr).name
                changelog = self.get_changelog_since_latest_stable_build(package, nvr)
                if changelog:
                    rendered_note += (
                        f"\n\n##### **Changelog for {package}**\n\n```\n"
                        f"{changelog}\n```"
                    )
            bugzilla_ids_from_changelog = self.get_bugzilla_ids_from_changelog(
                rendered_note,
            )
        try:
            save_kwargs = {
                "builds": koji_builds,
                "notes": rendered_note,
                "type": update_type,
            }

            bugs = []

            if bugzilla_ids:
                bugs = list(map(str, bugzilla_ids))
            if bugzilla_ids_from_changelog:
                bugs += bugzilla_ids_from_changelog

            if sidetag:
                save_kwargs["from_tag"] = sidetag

            if bugs:
                save_kwargs["bugs"] = bugs

            if alias:
                save_kwargs["edited"] = alias

            result = bodhi_client.save(**save_kwargs)

            logger.debug(f"Bodhi response:\n{result}")
            logger.info(
                f"Bodhi update {result['alias']}:\n"
                f"- {result['url']}\n"
                f"- stable_karma: {result['stable_karma']}\n"
                f"- unstable_karma: {result['unstable_karma']}\n"
                f"- notes:\n{result['notes']}\n",
            )
            if "caveats" in result:
                for cav in result["caveats"]:
                    logger.info(f"- {cav['name']}: {cav['description']}\n")

            return result["alias"], result["url"]

        except BodhiClientException as ex:
            # don't logger.error here as it will end in sentry and it may just
            # be a transient issue: e.g. waiting for a build to be tagged
            logger.info(ex)

            # early return in case update already exists
            if re.match(EXISTING_BODHI_UPDATE_REGEX, str(ex)):
                return None

            raise PackitBodhiException(
                f"There is a problem with creating a bodhi update:\n{ex}",
            ) from ex

    def get_allowed_gpg_keys_from_downstream_config(self) -> Optional[list[str]]:
        if not self.downstream_config:
            return None

        # SAFETY: Taking first package instead of specific package should be
        # safe, since we have put a requirement on »one« ‹upstream_project_url›
        # per Packit config, i.e. even if we're dealing with a monorepo, there
        # is only »one« upstream. If there is one upstream, there is only one
        # set of GPG keys that can be allowed.
        return self.downstream_config.packages[
            self.downstream_config._first_package
        ].allowed_gpg_keys

    def get_user(self) -> Optional[str]:
        return user if (user := super().get_user()) else self.fas_user

    @staticmethod
    def get_bugzilla_ids_from_changelog(changelog: str) -> list[str]:
        """
        Find bugzilla IDs referenced in the changelog.
        """
        # https://github.com/fedora-infra/bodhi/blob/373fe7f449e66a07ab0649c369bb68b11ec6d86f/bodhi-server/bodhi/server/config.py#L317
        bz_regex = (
            r"(?:fix(?:es)?|close(?:s)?|resolve(?:s)?)(?:\:)?\s"
            r"(?:fedora|epel|rh(?:bz)?)#(\d{5,})"
        )

        bugs = [bug.group(1) for bug in re.finditer(bz_regex, changelog, re.IGNORECASE)]
        logger.debug(f"Bug IDs found in the changelog: {bugs}")
        return bugs

    def get_nvr(self, dist_git_branch: str) -> str:
        """
        Get the NVR for the branch.

        Returns:
            nvr
        """
        self.create_branch(
            dist_git_branch,
            base=f"remotes/origin/{dist_git_branch}",
            setup_tracking=True,
        )

        self.update_branch(dist_git_branch)
        self.switch_branch(dist_git_branch)
        pkg_tool = PkgTool(
            fas_username=self.fas_user,
            directory=self.local_project.working_dir,
            tool=self.pkg_tool,
        )
        return pkg_tool.verrel()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from requre.cassette import DataTypes
from requre.helpers.files import StoreFiles
from requre.helpers.simple_object import Simple
from requre.modules_decorate_all_methods import (
    record_git_module,
    record_requests_module,
    record_tempfile_module,
)
from requre.online_replacing import (
    apply_decorator_to_all_methods,
    replace_module_match,
)

from packit.status import Status
from tests_recording.testbase import PackitTest


@record_tempfile_module
@record_requests_module
@record_git_module
@apply_decorator_to_all_methods(
    replace_module_match(
        what="packit.utils.run_command_remote",
        decorate=Simple.decorator_plain(),
    ),
)
@apply_decorator_to_all_methods(
    replace_module_match(
        what="packit.pkgtool.PkgTool.clone",
        decorate=StoreFiles.where_arg_references(
            key_position_params_dict={"target_path": 2},
        ),
    ),
)
@apply_decorator_to_all_methods(
    replace_module_match(
        what="copr.v3.helpers.config_from_file",
        decorate=Simple.decorator_plain(),
    ),
)
class TestStatus(PackitTest):
    def cassette_setup(self, cassette):
        cassette.data_miner.data_type = DataTypes.Dict

    def setUp(self):
        super().setUp()
        self._status = None

    @property
    def status(self):
        if not self._status:
            self._status = Status(self.config, self.pc, self.upstream, self.dg)
        return self._status

    def test_status(self):
        assert self.status

    def test_distgen_versions(self):
        table = self.status.get_dg_versions()
        assert table
        assert len(table) >= 3

    def test_koji_builds(self):
        table = self.status.get_koji_builds()
        assert table
        assert len(table) >= 2

    def test_copr_builds(self):
        table = self.status.get_copr_builds()
        assert table
        assert len(table) >= 2

    def test_up_releases(self):
        table = self.status.get_up_releases()
        assert len(table) >= 1

    def test_dowstream_pr(self):
        table = self.status.get_downstream_prs()
        assert len(table) >= 0



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy
import functools
import logging
import pathlib
import sys
from pathlib import Path
from typing import Optional, Union

import click
from github import GithubException
from ogr.parsing import parse_git_repo

from packit.api import PackitAPI
from packit.config import Config, JobType, get_local_package_config
from packit.config.common_package_config import MultiplePackages
from packit.config.package_config import PackageConfig
from packit.constants import CONFIG_FILE_NAMES, DISTRO_DIR, SRC_GIT_CONFIG
from packit.exceptions import PackitException, PackitNotAGitRepoException
from packit.local_project import LocalProject

logger = logging.getLogger(__name__)


def cover_packit_exception(_func=None, *, exit_code=None):
    """
    Decorator for executing the function in the try-except block.

    The PackitException is caught and
    - raised in debug
    - sys.exit(exit_code), otherwise

    On other Exceptions we print the message about creating an issue.


    If the function receives config, it recognises the debug mode.
    => use it after the @pass_config decorator
    """

    def decorator_cover(func):
        @functools.wraps(func)
        def covered_func(config=None, *args, **kwargs):
            try:
                if config:
                    func(config=config, *args, **kwargs)  # noqa: B026
                else:
                    func(*args, **kwargs)
            except KeyboardInterrupt:
                click.echo("Quitting on user request.")
                sys.exit(1)
            except PackitException as exc:
                if config and config.debug:
                    logger.exception(exc)
                else:
                    logger.error(exc)
                sys.exit(exit_code or 2)
            except GithubException as exc:
                if config and config.debug:
                    logger.exception(exc)
                else:
                    click.echo(
                        "We've encountered an error while talking to GitHub API, please make sure"
                        " that you pass GitHub API token and it has correct permissions, \n"
                        f"precise error message: {exc} \n"
                        "https://github.com/packit/packit/tree/master/docs\n",
                    )
                sys.exit(exit_code or 3)
            except Exception as exc:
                if config and config.debug:
                    logger.exception(exc)
                else:
                    logger.error(exc)
                    click.echo(
                        "Unexpected exception occurred,\n"
                        "please fill an issue here:\n"
                        "https://github.com/packit/packit/issues",
                        err=True,
                    )
                sys.exit(exit_code or 4)

        return covered_func

    return decorator_cover if _func is None else decorator_cover(_func)


def iterate_packages(func):
    """
    Decorator for dealing with sub-packages in a package (Monorepo) configuration

    * if packages are specified as an option in CLI then
      call decorated function just for them
    * if packages are not specified as an option in CLI but
      there are multiple packages in the configuration
      then call the decorated function for all of them
    * if there is just one package in the configuration
      then call the decorated function just once

    This method (iterate_packages) **has not** `package_config` key
    in its kwargs, it has `packages`, but calls a method
    (func) who needs a `package_config` key and not `packages`!
    """

    @functools.wraps(func)
    def covered_func(*args, **kwargs):
        path_or_url = kwargs["path_or_url"]
        config = kwargs["config"]
        decorated_func_kwargs = kwargs.copy()
        del decorated_func_kwargs["package"]
        packages_config: MultiplePackages = get_local_package_config(
            path_or_url.working_dir,
            repo_name=path_or_url.repo_name,
            try_local_dir_last=True,
            package_config_path=config.package_config_path,
        )
        packages_config_views_names = set(
            packages_config.get_package_config_views().keys(),
        )
        if kwargs.get("package"):
            if not_defined_packages := set(kwargs["package"]).difference(
                packages_config_views_names,
            ):
                logger.error(
                    "Packages %s are not defined in packit configuration.",
                    not_defined_packages,
                )
                return
            for package in kwargs["package"]:
                decorated_func_kwargs["config"] = copy.deepcopy(
                    config,
                )  # reset working variables like srpm_path
                decorated_func_kwargs[
                    "package_config"
                ] = packages_config.get_package_config_views()[package]
                func(*args, **decorated_func_kwargs)
        elif hasattr(packages_config, "packages"):
            for package_config in packages_config.get_package_config_views().values():
                decorated_func_kwargs["config"] = copy.deepcopy(
                    config,
                )  # reset working variables like srpm_path
                decorated_func_kwargs["package_config"] = package_config
                func(*args, **decorated_func_kwargs)
        else:
            logger.error("Given packages_config has no packages attribute")

    return covered_func


def iterate_packages_source_git(func):
    """
    Decorator for dealing with sub-packages in a package (Monorepo) configuration
    Designed for source-git related commands.

    * if dist-git-path is a git repo
      then search for a downstream repo name match
      in the configuration and find out its PackageConfig(View)
      but if there is just one package config and not matches
      then try going on with the found package config
    * if there are multiple package configs and
      dist-git-path is a dir (and not a git repo)
      for every sub-dir which is also a git repo
      search for a downstream repo name match
      in the configuration and find out its PackageConfig(View)
      Invoke the decorated function as many time as we
      found a match

    This method (iterate_packages) **has not** `package_config` key
    in its kwargs, but calls a method
    (func) who needs a `package_config` key!
    """

    @functools.wraps(func)
    def covered_func(*args, **kwargs):
        source_git = kwargs["source_git"]
        dist_git = kwargs["dist_git"]
        config = kwargs["config"]
        decorated_func_kwargs = kwargs.copy()

        source_git_path = pathlib.Path(source_git).resolve()
        dist_git_path = pathlib.Path(dist_git).resolve()

        packages_config = get_local_package_config(
            package_config_path=source_git_path / DISTRO_DIR / SRC_GIT_CONFIG,
        )

        found_func = False
        for package in packages_config.get_package_config_views().values():
            if package.downstream_package_name == dist_git_path.name:
                decorated_func_kwargs["config"] = copy.deepcopy(
                    config,
                )  # reset working variables like srpm_path
                decorated_func_kwargs["package_config"] = package
                func(*args, **decorated_func_kwargs)
                found_func = True

        # if names does not match but there is just one package try it
        if (
            len(packages_config.get_package_config_views()) == 1
            and dist_git_path.joinpath(".git").exists()
            and not found_func
        ):
            decorated_func_kwargs["config"] = copy.deepcopy(
                config,
            )  # reset working variables like srpm_path
            decorated_func_kwargs["package_config"] = next(
                iter(packages_config.get_package_config_views().values()),
            )
            func(*args, **decorated_func_kwargs)
            found_func = True

        # probably, if dist-git is not a git repo,
        # we would like to cycle over multiple dist-git repos
        elif (
            dist_git_path.is_dir()
            and not dist_git_path.joinpath(".git").exists()
            and not found_func
        ):
            repo_dirs = [
                p
                for p in dist_git_path.glob("*")
                if p.is_dir() and p.joinpath(".git").exists()
            ]
            for package in packages_config.get_package_config_views().values():
                for repo_dir in repo_dirs:
                    if package.downstream_package_name == repo_dir.name:
                        decorated_func_kwargs["config"] = copy.deepcopy(
                            config,
                        )  # reset working variables like srpm_path
                        decorated_func_kwargs["dist_git"] = str(repo_dir)
                        decorated_func_kwargs["package_config"] = package
                        try:
                            func(*args, **decorated_func_kwargs)
                        except PackitException as ex:
                            if "git trailer does not exist" in str(ex):
                                pass
                            else:
                                raise ex
                        found_func = True

        if not found_func:
            logger.error(
                f"No match found for source git {source_git} and dist git {dist_git}.",
            )

    return covered_func


def get_packit_api(
    config: Config,
    local_project: LocalProject,
    package_config: Optional[Union[PackageConfig, MultiplePackages]] = None,
    dist_git_path: Optional[str] = None,
    job_config_index: Optional[int] = None,
    job_type: Optional[JobType] = None,
    check_for_non_git_upstream: Optional[bool] = False,
) -> PackitAPI:
    """
    Load the package config, set other options and return the PackitAPI
    """
    if not package_config:
        # TODO: to be removed when monorepo refactoring is finished!
        package_config = get_local_package_config(
            local_project.working_dir,
            repo_name=local_project.repo_name,
            try_local_dir_last=True,
            package_config_path=config.package_config_path,
        )

    if check_for_non_git_upstream and package_config.upstream_project_url is None:
        logger.debug("Upstream will be treated as non-git.")
        non_git_upstream = True
    else:
        non_git_upstream = False

    logger.debug(f"job_config_index: {job_config_index}")
    if job_config_index is not None and isinstance(package_config, PackageConfig):
        if job_config_index >= len(package_config.jobs):
            raise PackitException(
                "job_config_index is bigger than number of jobs in package config!",
            )
        package_config = package_config.jobs[job_config_index]
        logger.debug(f"Final package (job) config: {package_config}")
    elif job_type is not None:
        try:
            package_config = next(
                job for job in package_config.jobs if job.type == job_type
            )
        except IndexError as e:
            raise PackitException(
                f"No job with type {job_type} found in package config.",
            ) from e
        logger.debug(f"Final package (job) config: {package_config}")

    if dist_git_path and Path(dist_git_path) == local_project.working_dir:
        return PackitAPI(
            config=config,
            package_config=package_config,
            upstream_local_project=None,
            downstream_local_project=local_project,
            dist_git_clone_path=dist_git_path,
            non_git_upstream=non_git_upstream,
        )

    if not local_project.git_repo:
        raise PackitNotAGitRepoException(
            f"{local_project.working_dir!r} is not a git repository.",
        )

    remote_urls: list[str] = []
    for remote in local_project.git_repo.remotes:
        remote_urls += remote.urls

    upstream_hostname = (
        get_hostname_or_none(url=package_config.upstream_project_url)
        if package_config.upstream_project_url
        else None
    )

    lp_upstream = None
    lp_downstream = None

    for url in remote_urls:
        parsed_url = parse_git_repo(url)
        if not parsed_url.hostname:
            continue

        if package_config.dist_git_instance.has_repository(url):
            lp_downstream = local_project
            logger.debug(
                "Input directory is a downstream repository. Deduced from package config.",
            )
            break

        if upstream_hostname and parsed_url.hostname == upstream_hostname:
            lp_upstream = local_project
            logger.debug(
                "Input directory is an upstream repository. Upstream hostname matches.",
            )
            break
    else:
        lp_upstream = local_project
        # fallback, this is the past behavior
        logger.debug(
            "Input directory is an upstream repository. No suitable remote found.",
        )

    return PackitAPI(
        config=config,
        package_config=package_config,
        upstream_local_project=lp_upstream,
        downstream_local_project=lp_downstream,
        dist_git_clone_path=dist_git_path,
        non_git_upstream=non_git_upstream,
    )


def get_hostname_or_none(url: str) -> Optional[str]:
    parsed_url = parse_git_repo(potential_url=url)
    if parsed_url:
        return parsed_url.hostname
    return None


def get_existing_config(working_dir: Path) -> Optional[Path]:
    # find name of config file if already exists
    for config_file_name in CONFIG_FILE_NAMES:
        config_file_path = working_dir / config_file_name
        if config_file_path.is_file():
            return config_file_path
    return None



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Update selected component from upstream in Fedora
"""

import functools
import logging
import os

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import PackageConfig, get_context_settings, pass_config
from packit.config.aliases import get_branches, get_fast_forward_merge_branches_for
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)

logger = logging.getLogger(__name__)


def get_dist_git_branches(api, dist_git_branch, pull_from_upstream=False):
    cmdline_dg_branches = dist_git_branch.split(",") if dist_git_branch else []
    config_dg_branches = []
    if isinstance(api.package_config, PackageConfig):
        config_dg_branches = (
            api.package_config.get_propose_downstream_dg_branches_value(
                pull_from_upstream=pull_from_upstream,
            )
        )

    default_dg_branch = api.dg.local_project.git_project.default_branch

    dg_branches = (
        cmdline_dg_branches or config_dg_branches or default_dg_branch.split(",")
    )
    return dg_branches, default_dg_branch


def get_dg_branches(api, dist_git_branch, pull_from_upstream=False):
    dg_branches, default_dg_branch = get_dist_git_branches(
        api,
        dist_git_branch,
        pull_from_upstream,
    )
    return get_branches(*dg_branches, default_dg_branch=default_dg_branch)


def sync_release(
    config,
    dist_git_path,
    dist_git_branch,
    force_new_sources,
    pr,
    local_content,
    path_or_url,
    upstream_ref,
    version,
    force,
    use_downstream_specfile,
    package_config,
    resolve_bug,
    sync_acls,
    check_for_non_git_upstream=False,
):
    api = get_packit_api(
        config=config,
        package_config=package_config,
        dist_git_path=dist_git_path,
        local_project=path_or_url,
        check_for_non_git_upstream=check_for_non_git_upstream,
    )
    if pr is None:
        pr = api.package_config.create_pr

    branches_to_update = get_dg_branches(
        api,
        dist_git_branch,
        pull_from_upstream=use_downstream_specfile,
    )

    click.echo(
        f"Proposing update of the following branches: {', '.join(branches_to_update)}",
    )

    dist_git_branches, default_dg_branch = get_dist_git_branches(
        api,
        dist_git_branch,
        pull_from_upstream=use_downstream_specfile,
    )
    branches_to_update = get_branches(
        *dist_git_branches,
        default_dg_branch=default_dg_branch,
    )
    for branch in branches_to_update:
        api.sync_release(
            dist_git_branch=branch,
            use_local_content=local_content,
            versions=[version] if version else [],
            force_new_sources=force_new_sources,
            upstream_ref=upstream_ref,
            create_pr=pr,
            force=force,
            use_downstream_specfile=use_downstream_specfile,
            resolve_bug=resolve_bug,
            sync_acls=sync_acls,
            fast_forward_merge_branches=get_fast_forward_merge_branches_for(
                dist_git_branches=dist_git_branches,
                source_branch=branch,
                default=default_dg_branch,
            ),
        )


def sync_release_common_options(func):
    @click.option(
        "--dist-git-branch",
        help="Comma separated list of target branches in dist-git to release into. "
        "(defaults to all branches)",
    )
    @click.option(
        "--dist-git-path",
        help="Path to dist-git repo to work in. "
        "Otherwise clone the repo in a temporary directory.",
    )
    @click.option(
        "--force-new-sources",
        is_flag=True,
        default=False,
        help="Upload the new sources also when the archive is already in the lookaside cache.",
    )
    @click.option(
        "--pr/--no-pr",
        default=None,
        help=(
            "Create a pull request to downstream repository or push directly. "
            "If not set, defaults to value set in configuration."
        ),
    )
    @click.option(
        "--force",
        "-f",
        default=False,
        is_flag=True,
        help="Don't discard changes in the git repo by default, unless this is set.",
    )
    @click.option(
        "-b",
        "--resolve-bug",
        help="Bug(s) that are resolved with the update, e.g. rhbz#123 (multiple can be specified)",
        required=False,
        multiple=True,
        type=click.INT,
    )
    @click.option(
        "--sync-acls",
        default=False,
        is_flag=True,
        help="Sync ACLs between dist-git repo and the fork, is considered only with --pr option.",
    )
    @click.option(
        PACKAGE_SHORT_OPTION,
        PACKAGE_LONG_OPTION,
        multiple=True,
        help=PACKAGE_OPTION_HELP.format(action="sync downstream"),
    )
    @click.argument(
        "path_or_url",
        type=LocalProjectParameter(),
        default=os.path.curdir,
    )
    @click.argument("version", required=False)
    @functools.wraps(func)
    def wrapper_common_options(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper_common_options


@click.command("propose-downstream", context_settings=get_context_settings())
@sync_release_common_options
@click.option(
    "--local-content",
    is_flag=True,
    default=False,
    help="Do not checkout release tag. Use the current state of the repo. "
    "This option is set by default for source-git repos",
)
@click.option(
    "--upstream-ref",
    default=None,
    help="Git ref of the last upstream commit in the current branch "
    "from which packit should generate patches "
    "(this option implies the repository is source-git).",
)
@pass_config
@cover_packit_exception
@iterate_packages
def propose_downstream(
    config,
    dist_git_path,
    dist_git_branch,
    force_new_sources,
    pr,
    path_or_url,
    version,
    local_content,
    upstream_ref,
    force,
    sync_acls,
    resolve_bug,
    package_config,
):
    """
    Land a new upstream release in Fedora using upstream packit config.

    PATH_OR_URL argument is a local path or a URL to the upstream git repository,
    it defaults to the current working directory

    VERSION argument is optional, the latest upstream version
    will be used by default
    """
    sync_release(
        config=config,
        dist_git_path=dist_git_path,
        dist_git_branch=dist_git_branch,
        force_new_sources=force_new_sources,
        pr=pr,
        path_or_url=path_or_url,
        version=version,
        force=force,
        local_content=local_content,
        upstream_ref=upstream_ref,
        use_downstream_specfile=False,
        package_config=package_config,
        resolve_bug=resolve_bug,
        sync_acls=sync_acls,
    )


@click.command("pull-from-upstream", context_settings=get_context_settings())
@sync_release_common_options
@pass_config
@cover_packit_exception
@iterate_packages
def pull_from_upstream(
    config,
    dist_git_path,
    dist_git_branch,
    force_new_sources,
    pr,
    path_or_url,
    version,
    force,
    sync_acls,
    resolve_bug,
    package_config,
):
    """
    Land a new upstream release in Fedora using downstream packit config.

    PATH_OR_URL argument is a local path or a URL to the dist-git repository,
    it defaults to the current working directory

    VERSION argument is optional, the latest upstream version
    will be used by default
    """
    sync_release(
        config=config,
        dist_git_path=dist_git_path,
        dist_git_branch=dist_git_branch,
        force_new_sources=force_new_sources,
        pr=pr,
        path_or_url=path_or_url,
        version=version,
        force=force,
        local_content=False,
        upstream_ref=None,
        use_downstream_specfile=True,
        package_config=package_config,
        resolve_bug=resolve_bug,
        sync_acls=sync_acls,
        check_for_non_git_upstream=True,
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

from packit.exceptions import PackitException
from packit.upstream import GitUpstream


def test_synch_push_and_up_repo_dirty(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_sync_push,
):
    """Check that exception is raised when upstream repo is dirty"""

    sourcegit, _ = sourcegit_and_remote
    (sourcegit / "README").write_text("a change in the repo")

    with pytest.raises(PackitException) as exc:
        api_instance_sync_push.sync_push()

    assert "is dirty" in str(exc.value)


def test_synch_push_and_dg_repo_dirty(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_sync_push,
):
    """Check that exception is raised when dist-git repo is dirty"""
    distgit, _ = distgit_and_remote
    (distgit / "README").write_text("a change in the repo")

    with pytest.raises(PackitException) as exc:
        api_instance_sync_push.sync_push()

    assert "is dirty" in str(exc.value)


def test_synch_push_and_diverged_repos(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_sync_push,
):
    """Check that exception is raised when upstream and dist-git repos have diverged"""

    api_instance_sync_push.up.specfile.version = "2.3.4"
    api_instance_sync_push.up.commit("Source-git commit to be synced", "")

    with pytest.raises(PackitException) as exc:
        api_instance_sync_push.sync_push()

    assert "diverged" in str(exc.value)


def test_synch_push_one_commit(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_sync_push,
):
    """Update upstream if dist-git has a new commit"""

    flexmock(GitUpstream).should_receive("push_to_fork").and_return(
        "main",
        "packit.dev",
    ).once()
    flexmock(GitUpstream).should_receive("create_pull").and_return(None).once()

    assert (
        "dist-git commit to be sync back"
        not in api_instance_sync_push.up.local_project.git_repo.head.commit.message
    )

    api_instance_sync_push.sync_push()

    assert (
        "dist-git commit to be sync back"
        in api_instance_sync_push.up.local_project.git_repo.head.commit.message
    )


def test_synch_push_two_commits(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_sync_push,
):
    """Update upstream if dist-git has a new commit"""
    api_instance_sync_push.dg.specfile.version = "3.4.5"
    api_instance_sync_push.dg.commit(
        "Another dist-git commit to be synced back\n\nWith a multiline\ndescription",
        "",
    )

    flexmock(GitUpstream).should_receive("push_to_fork").and_return(
        "main",
        "packit.dev",
    ).once()
    flexmock(GitUpstream).should_receive("create_pull").and_return(None).once()

    commits = api_instance_sync_push.up.local_project.git_repo.iter_commits()
    commits = [c.summary for c in commits]
    assert "dist-git commit to be sync back" not in commits
    assert "Another dist-git commit to be synced back" not in commits

    api_instance_sync_push.sync_push()

    commits = api_instance_sync_push.up.local_project.git_repo.iter_commits()
    commits = [c.summary for c in commits]
    assert "[packit] dist-git commit to be sync back" in commits
    assert "[packit] Another dist-git commit to be synced back" in commits



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

from packit.api import PackitAPI
from packit.cli.propose_downstream import get_dg_branches
from packit.config import (
    CommonPackageConfig,
    JobConfig,
    JobConfigTriggerType,
    JobType,
    PackageConfig,
)


@pytest.mark.parametrize(
    "package_config,cmdline,expected",
    [
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                            ),
                        },
                    ),
                ],
            ),
            None,
            {
                "default_br",
            },
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                dist_git_branches=["file"],
                            ),
                        },
                    ),
                ],
            ),
            None,
            {
                "file",
            },
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                            ),
                        },
                    ),
                ],
            ),
            "cmdline",
            {
                "cmdline",
            },
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                dist_git_branches=["file"],
                            ),
                        },
                    ),
                ],
            ),
            "cmdline",
            {
                "cmdline",
            },
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                dist_git_branches=["file1", "file2"],
                            ),
                        },
                    ),
                ],
            ),
            None,
            {
                "file1",
                "file2",
            },
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                            ),
                        },
                    ),
                ],
            ),
            "cmdline1,cmdline2",
            {
                "cmdline1",
                "cmdline2",
            },
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                            ),
                        },
                    ),
                ],
            ),
            "rawhide",
            {
                "default_br",
            },
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                dist_git_branches=["rawhide"],
                            ),
                        },
                    ),
                ],
            ),
            None,
            {
                "default_br",
            },
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                dist_git_branches=["file1", "rawhide"],
                            ),
                        },
                    ),
                ],
            ),
            None,
            {
                "file1",
                "default_br",
            },
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                            ),
                        },
                    ),
                ],
            ),
            "cmdline1,rawhide",
            {
                "cmdline1",
                "default_br",
            },
        ),
    ],
)
def test_get_dist_git_branches(package_config, cmdline, expected):
    api = flexmock(PackitAPI)
    api.package_config = package_config
    git_project = flexmock(default_branch="default_br")
    local_project = flexmock(git_project=git_project)
    api.dg = flexmock(local_project=local_project)

    assert get_dg_branches(api, cmdline) == expected



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""The 'source-git' subcommand for Packit"""

import click

from packit.cli.source_git_init import source_git_init
from packit.cli.source_git_status import source_git_status
from packit.cli.update_dist_git import update_dist_git
from packit.cli.update_source_git import update_source_git


@click.group("source-git")
def source_git():
    """Subcommand to collect source-git related functionality"""


source_git.add_command(update_dist_git)
source_git.add_command(update_source_git)
source_git.add_command(source_git_init)
source_git.add_command(source_git_status)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pathlib
from contextlib import suppress as does_not_raise

import pytest
from bugzilla import Bugzilla
from flexmock import flexmock
from munch import Munch

from packit import api
from packit import api as packit_api
from packit.api import PackitAPI
from packit.config import CommonPackageConfig, PackageConfig, RunCommandType
from packit.config.config import Config
from packit.copr_helper import CoprHelper
from packit.distgit import DistGit
from packit.exceptions import PackitException, ReleaseSkippedPackitException
from packit.local_project import LocalProjectBuilder
from packit.patches import PatchGenerator
from packit.sync import SyncFilesItem
from packit.utils.changelog_helper import ChangelogHelper


def build_dict(copr_url, id):
    """Create a build object which uses 'copr_url' and 'id'."""
    # copr_client.build_proxy.get(build_id) response
    return Munch(
        {
            "chroots": [
                "fedora-29-x86_64",
                "fedora-30-x86_64",
                "fedora-rawhide-x86_64",
            ],
            "ended_on": 1566377991,
            "id": str(id),
            "ownername": "packit",
            "project_dirname": "packit-service-ogr-160",
            "projectname": "packit-service-ogr-160",
            "repo_url": f"{copr_url}/results/packit/packit-service-ogr-160",
            "source_package": {
                "name": "python-ogr",
                "url": "https://copr-be.cloud.fedoraproject.org/results/"
                "packit/packit-service-ogr-160/srpm-builds/01010428/"
                "python-ogr-0.6.1.dev51ge88ac83-1.fc30.src.rpm",
                "version": "0.6.1.dev51+ge88ac83-1.fc30",
            },
            "started_on": 1566377844,
            "state": "succeeded",
            "submitted_on": 1566377764,
            "submitter": "packit",
        },
    )


def copr_helper(copr_url):
    """Create a mock CoprHelper, with a copr_client configured with 'copr_url'."""
    helper = CoprHelper(flexmock())
    helper._copr_client = flexmock(config={"copr_url": copr_url})
    return helper


testdata = [
    pytest.param(
        copr_helper("https://supr.copr"),
        build_dict("https://supr.copr", 1010428),
        "https://supr.copr/coprs/build/1010428/",
        id="user",
    ),
    # The name "group" bellow is kept for historical reasons.
    # These Copr permalinks have no information in them regarding who
    # the owner of the build is (although they will have, once they redirect).
    pytest.param(
        copr_helper("https://group.copr"),
        build_dict("https://group.copr", 1010430),
        "https://group.copr/coprs/build/1010430/",
        id="group",
    ),
]


@pytest.fixture
def api_mock(config_mock, package_config_mock, upstream_mock, distgit_mock):
    api = PackitAPI(config=config_mock, package_config=package_config_mock)
    flexmock(api)
    api._up = upstream_mock
    api._dg = distgit_mock
    api.should_receive("_prepare_files_to_sync").and_return([])
    api.should_receive("_handle_sources")
    api.should_receive("_get_sandcastle_exec_dir").and_return("sandcastle-exec-dir")
    return api


@pytest.mark.parametrize(
    "helper,build,web_url",
    testdata,
)
class TestPackitAPI:
    def test_copr_web_build_url(self, helper, build, web_url):
        assert helper.copr_web_build_url(build) == web_url


@pytest.mark.parametrize(
    "version, tag, get_latest_released_return, get_specfile_version_return, expectation",
    [
        pytest.param(
            "1.1.1",
            None,
            None,
            "0",
            does_not_raise(),
            id="version_set(CLI_explicit)",
        ),
        pytest.param(
            None,
            "v1.1.1",
            None,
            "0",
            does_not_raise(),
            id="tag_set(service_mode)",
        ),
        pytest.param(
            "1.1",
            "v1.1.1",
            None,
            None,
            pytest.raises(PackitException),
            id="both_set(CLI_wrong_usage)",
        ),
        pytest.param(
            None,
            None,
            "1.1",
            "0",
            does_not_raise(),
            id="none_set(CLI_version_from_upstream_release_monitoring)",
        ),
        pytest.param(
            None,
            None,
            None,
            None,
            pytest.raises(PackitException),
            id="none_set(CLI_version_not_in_upstream_release_monitoring)",
        ),
    ],
)
def test_sync_release_version_tag_processing(
    version,
    tag,
    get_latest_released_return,
    get_specfile_version_return,
    expectation,
    api_mock,
):
    api_mock.up.package_config.upstream_tag_template = "v{version}"
    api_mock.up.should_receive("get_latest_released_version").and_return(
        get_latest_released_return,
    )
    api_mock.up.should_receive("get_specfile_version").and_return(
        get_specfile_version_return,
    )
    api_mock.up.should_receive("specfile").and_return(
        flexmock().should_receive("reload").mock(),
    )
    api_mock.up.package_config.should_receive("get_package_names_as_env").and_return({})
    api_mock.should_receive("_prepare_files_to_sync").with_args(
        synced_files=[],
        full_version=version,
        upstream_tag=tag,
    )
    api_mock.should_receive("push_and_create_pr").and_return(flexmock())
    flexmock(PatchGenerator).should_receive("undo_identical")
    versions = [version] if version else []
    with expectation:
        api_mock.sync_release(versions=versions, tag=tag, dist_git_branch="_")


def test_sync_release_do_not_create_sync_note(api_mock):
    flexmock(PatchGenerator).should_receive("undo_identical")
    flexmock(pathlib.Path).should_receive("write_text").never()
    api_mock.up.should_receive("get_specfile_version").and_return("0")
    api_mock.up.should_receive("specfile").and_return(
        flexmock().should_receive("reload").mock(),
    )
    api_mock.up.package_config.should_receive("get_package_names_as_env").and_return({})
    api_mock.up.package_config.create_sync_note = False
    api_mock.should_receive("push_and_create_pr").and_return(flexmock())
    api_mock.sync_release(versions=["1.1"], dist_git_branch="_")


def test_sync_release_create_sync_note(api_mock):
    flexmock(PatchGenerator).should_receive("undo_identical")
    flexmock(pathlib.Path).should_receive("write_text").once()
    api_mock.up.should_receive("get_specfile_version").and_return("0")
    api_mock.up.should_receive("specfile").and_return(
        flexmock().should_receive("reload").mock(),
    )
    api_mock.up.package_config.should_receive("get_package_names_as_env").and_return({})
    api_mock.should_receive("push_and_create_pr").and_return(flexmock())
    api_mock.sync_release(versions=["1.1"], dist_git_branch="_")


def test_sync_release_warn_about_koji_build_triggering_bug(api_mock):
    flexmock(PatchGenerator).should_receive("undo_identical")
    flexmock(pathlib.Path).should_receive("write_text").once()
    api_mock.up.should_receive("get_specfile_version").and_return("0")
    api_mock.up.should_receive("specfile").and_return(
        flexmock().should_receive("reload").mock(),
    )
    api_mock.up.package_config.should_receive("get_package_names_as_env").and_return({})
    api_mock.dg.should_receive("push_to_fork").and_return()
    pr = flexmock(
        _raw_pr={"commit_start": "1234abc", "commit_stop": "5678def"},
        target_branch="_",
    )
    api_mock.dg.should_receive("create_pull").and_return(pr)
    flexmock(api).should_receive("get_branches").and_return({"_", "__"})
    pr.should_receive("comment").once()
    api_mock.sync_release(
        versions=["1.1"],
        dist_git_branch="_",
        warn_about_koji_build_triggering_bug=True,
    )


def test_common_env(api_mock):
    env = api_mock.common_env()
    assert env == {
        "PACKIT_DOWNSTREAM_REPO": "/mock_dir/sandcastle/dist-git",
        "PACKIT_UPSTREAM_REPO": "/mock_dir/sandcastle/local-project",
        "PACKIT_PWD": "/mock_dir/sandcastle/local-project",
    }
    api_mock.config.command_handler = RunCommandType.sandcastle
    env = api_mock.common_env()
    assert env == {
        "PACKIT_DOWNSTREAM_REPO": "/mock_dir/sandcastle/sandcastle-exec-dir/dist-git",
        "PACKIT_UPSTREAM_REPO": "/mock_dir/sandcastle/sandcastle-exec-dir/local-project",
        "PACKIT_PWD": "/mock_dir/sandcastle/sandcastle-exec-dir/local-project",
    }


@pytest.mark.parametrize(
    "path, downstream_package_name, expectation",
    [
        pytest.param("/systemd", "systemd", "systemd", id="both_set"),
        pytest.param(None, "systemd", "systemd", id="both_set"),
        pytest.param("/systemd", None, "systemd", id="both_set"),
        pytest.param(None, None, None, id="none_set"),
    ],
)
def test_dg_downstream_package_name_is_set(
    api_mock,
    path,
    downstream_package_name,
    expectation,
):
    api_mock._dg = None
    api_mock.package_config.downstream_package_name = downstream_package_name
    api_mock.downstream_local_project = LocalProjectBuilder().build(working_dir=path)
    assert api_mock.dg.package_config.downstream_package_name == expectation


def test_sync_release_sync_files_call(config_mock, upstream_mock, distgit_mock):
    pc = PackageConfig(
        packages={
            "package": CommonPackageConfig(
                specfile_path="xxx",
                files_to_sync=[
                    SyncFilesItem(
                        ["src/a"],
                        "dest",
                        filters=["dummy filter"],
                        mkpath=True,
                        delete=True,
                    ),
                ],
                upstream_package_name="test_package_name",
                downstream_package_name="test_package_name",
                upstream_tag_template="_",
                upstream_project_url="_",
                upstream_ref="_",
            ),
        },
    )
    upstream_mock.package_config = pc
    distgit_mock.package_config = pc
    api = PackitAPI(config=config_mock, package_config=pc)
    flexmock(api)
    api._up = upstream_mock
    api._dg = distgit_mock
    api.should_receive("_handle_sources")
    flexmock(PatchGenerator).should_receive("undo_identical")
    flexmock(pathlib.Path).should_receive("write_text").once()
    api.up.should_receive("get_specfile_version").and_return("0")
    api.up.should_receive("specfile").and_return(
        flexmock().should_receive("reload").mock(),
    )
    api.should_receive("push_and_create_pr").and_return(flexmock())
    flexmock(ChangelogHelper).should_receive("update_dist_git")

    flexmock(packit_api).should_receive("sync_files").with_args(
        [
            SyncFilesItem(
                src=["/mock_dir/sandcastle/local-project/src/a"],
                dest="/mock_dir/sandcastle/dist-git/dest",
                mkpath=True,
                delete=True,
                filters=["dummy filter"],
            ),
        ],
    )

    api.sync_release(versions=["1.1"], dist_git_branch="_")


def test_sync_release_check_pr_instructions(api_mock):
    flexmock(PatchGenerator).should_receive("undo_identical")
    api_mock.up.should_receive("get_specfile_version").and_return("0")
    api_mock.up.should_receive("specfile").and_return(
        flexmock().should_receive("reload").mock(),
    )
    api_mock.up.package_config.should_receive("get_package_names_as_env").and_return({})
    api_mock.should_receive("push_and_create_pr").with_args(
        pr_title=str,
        pr_description=(
            "Upstream tag: _\nUpstream commit: _\n\n---\n\n"
            "If you need to do any change in this pull request, you can clone Packit's fork "
            "and push directly to the source branch of this PR (provided you have commit access "
            "to this repository):\n"
            "```\n"
            "git clone ssh://$YOUR_USER@pkgs.fedoraproject.org/forks/packit/rpms/package.git\n"
            "cd package\n"
            "git checkout _-update\n"
            "git push origin _-update\n"
            "```\n"
            "\n---\n\n"
            "Alternatively, if you already have the package repository cloned, "
            "you can just fetch the Packit's fork:\n"
            "```\n"
            "cd package\n"
            "git remote add packit ssh://$YOUR_USER@pkgs.fedoraproject.org/forks/packit/rpms/package.git\n"
            "git fetch packit refs/heads/_-update\n"
            "git checkout _-update\n"
            "git push packit _-update\n"
            "```\n"
            "If you have the `koji_build` job configured as well, make sure to configure "
            "the `allowed_pr_authors` and/or `allowed_committers` (see [the docs]"
            "(https://packit.dev/docs/configuration/downstream/koji_build#optional-parameters)) "
            "since by default, Packit reacts only to its own PRs.\n"
            "\n---\n\n"
            "Before pushing builds/updates, please remember to check the new "
            "version against the "
            "[packaging guidelines](https://docs.fedoraproject.org/en-US/packaging-guidelines/).\n\n"
            "For example, please:\n\n"
            "- check that the new sources only contain permissible content\n"
            "- check that the license of the new version has not changed\n"
            "- check for any API/ABI and other changes that may break dependent packages\n"
            "- check the autogenerated changelog\n"
        ),
        git_branch=str,
        repo=DistGit,
        sync_acls=False,
    ).and_return(flexmock())
    api_mock.sync_release(
        versions=["1.1"],
        dist_git_branch="_",
        add_pr_instructions=True,
    )


def test_sync_release_downgrade(api_mock):
    flexmock(PatchGenerator).should_receive("undo_identical")
    api_mock.up.should_receive("get_specfile_version").and_return("0")
    api_mock.up.should_receive("specfile").and_return(
        flexmock().should_receive("reload").mock(),
    )
    api_mock.up.package_config.should_receive("get_package_names_as_env").and_return({})
    api_mock.dg.should_receive("get_specfile_version").and_return("1.1")
    with pytest.raises(ReleaseSkippedPackitException):
        api_mock.sync_release(
            versions=["1.0"],
            dist_git_branch="_",
            add_pr_instructions=True,
        )


@pytest.mark.parametrize(
    "resolved_bugs, result",
    [
        pytest.param(
            ["rhbz#123"],
            "- Resolves: rhbz#123\n\nUpstream tag: 1.0.0\nUpstream commit: _\n"
            "\nCommit authored by Packit automation (https://packit.dev/)\n",
        ),
        pytest.param(
            ["rhbz#123", "rhbz#222"],
            (
                "- Resolves: rhbz#123\n- Resolves: rhbz#222\n\n"
                "Upstream tag: 1.0.0\nUpstream commit: _\n"
                "\nCommit authored by Packit automation (https://packit.dev/)\n"
            ),
        ),
        pytest.param(
            None,
            "Upstream tag: 1.0.0\nUpstream commit: _\n"
            "\nCommit authored by Packit automation (https://packit.dev/)\n",
        ),
    ],
)
def test_get_default_commit_description(api_mock, resolved_bugs, result):
    assert (
        api_mock.get_default_commit_description("1.0.0", resolved_bugs=resolved_bugs)
        == result
    )


@pytest.mark.parametrize(
    "tag_link, commit_link, project_id, resolved_bugs, result",
    [
        pytest.param(
            "",
            "",
            None,
            None,
            "Upstream tag: 1.0.0\nUpstream commit: _\n",
        ),
        pytest.param(
            "tag-link",
            "",
            None,
            None,
            "Upstream tag: [1.0.0](tag-link)\nUpstream commit: _\n",
        ),
        pytest.param(
            "tag-link",
            "commit-link",
            None,
            None,
            "Upstream tag: [1.0.0](tag-link)\nUpstream commit: [_](commit-link)\n",
        ),
        pytest.param(
            "tag-link",
            "",
            None,
            None,
            "Upstream tag: [1.0.0](tag-link)\n" "Upstream commit: _\n",
        ),
        pytest.param(
            "tag-link",
            "commit-link",
            None,
            None,
            "Upstream tag: [1.0.0](tag-link)\n" "Upstream commit: [_](commit-link)\n",
        ),
        pytest.param(
            "tag-link",
            "commit-link",
            12345,
            None,
            "Upstream tag: [1.0.0](tag-link)\n"
            "Upstream commit: [_](commit-link)\n"
            "Release monitoring project: [12345](https://release-monitoring.org/project/12345)\n",
        ),
        pytest.param(
            "tag-link",
            "commit-link",
            12345,
            ["rhbz#1234"],
            "Upstream tag: [1.0.0](tag-link)\n"
            "Upstream commit: [_](commit-link)\n"
            "Release monitoring project: [12345](https://release-monitoring.org/project/12345)\n"
            "Resolves: [rhbz#1234](https://bugzilla.redhat.com/show_bug.cgi?id=1234)\n",
        ),
        pytest.param(
            "tag-link",
            "commit-link",
            12345,
            ["rhbz#not-a-number"],
            "Upstream tag: [1.0.0](tag-link)\n"
            "Upstream commit: [_](commit-link)\n"
            "Release monitoring project: [12345](https://release-monitoring.org/project/12345)\n"
            "Resolves: rhbz#not-a-number\n",
        ),
    ],
)
def test_get_pr_description(
    api_mock,
    tag_link,
    commit_link,
    project_id,
    resolved_bugs,
    result,
):
    flexmock(api).should_receive("get_tag_link").and_return(tag_link)
    flexmock(api).should_receive("get_commit_link").and_return(commit_link)
    assert (
        api_mock.get_pr_description(
            "1.0.0",
            release_monitoring_project_id=project_id,
            resolved_bugs=resolved_bugs,
        )
        == result
    )


@pytest.mark.parametrize(
    "package_config, config, expected_pkg_tool",
    (
        pytest.param(
            flexmock(pkg_tool=None),
            Config(),
            "fedpkg",
            id="default from config",
        ),
        pytest.param(
            flexmock(pkg_tool="rhpkg"),
            Config(),
            "rhpkg",
            id="package-level override",
        ),
        # regression in automated allowlisting
        pytest.param(None, Config(), "fedpkg", id="no package_config given"),
    ),
)
def test_pkg_tool_property(package_config, config, expected_pkg_tool):
    assert PackitAPI(config, package_config).pkg_tool == expected_pkg_tool


@pytest.mark.parametrize(
    "current_version, proposed_version, target_branch, version_update_mask, exp",
    (
        pytest.param(
            "3.10.0",
            "4.0.0",
            "rawhide",
            None,
            True,
            id="skip version distance check for rawhide",
        ),
        pytest.param(
            "3.10.0",
            "4.0.0",
            "f38",
            r"\d+\.\d+\.",
            False,
            id="proposed version far too distant for f38",
        ),
        pytest.param(
            "3.10.0",
            "3.10.1",
            "f38",
            r"\d+\.\d+\.",
            True,
            id="proposed version ok for f38",
        ),
    ),
)
def test_check_version_distance(
    current_version,
    proposed_version,
    target_branch,
    version_update_mask,
    exp,
):
    flexmock(api).should_receive("get_branches").and_return(["f38"])
    package_config = flexmock(
        version_update_mask=version_update_mask,
    )
    config = Config()

    assert (
        PackitAPI(config, package_config).check_version_distance(
            current_version,
            proposed_version,
            target_branch,
        )
        == exp
    )


@pytest.mark.parametrize(
    "package_name, version, response, result",
    (
        pytest.param(
            "python-ogr",
            "1.0.0",
            [
                flexmock(id=1, summary="python-ogr-1.1.1 is available"),
                flexmock(id=2, summary="python-ogr-1.0.0 is available"),
            ],
            "rhbz#2",
        ),
        pytest.param(
            "python-ogr",
            "2.0.0",
            [
                flexmock(id=1, summary="python-ogr-1.1.1 is available"),
                flexmock(id=2, summary="python-ogr-1.0.0 is available"),
            ],
            None,
        ),
        pytest.param(
            "python-ogr",
            "2.0.0",
            [
                flexmock(id=1, summary="python-ogr-1.1.1 is available"),
                flexmock(id=2, summary="python-ogr-1.0.0 is available"),
                flexmock(id=2, summary="python-ogr-2.0 is available"),
            ],
            None,
        ),
    ),
)
def test_get_upstream_release_monitoring_bug(package_name, version, response, result):
    flexmock(Bugzilla).should_receive("query").and_return(response)
    assert (
        PackitAPI.get_upstream_release_monitoring_bug(package_name, version) == result
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import contextlib
from importlib.metadata import PackageNotFoundError, version

with contextlib.suppress(PackageNotFoundError):
    __version__ = version("packitos")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import Optional

from packit.constants import COPR_SOURCE_SCRIPT


def create_source_script(
    url: str,
    ref: Optional[str] = None,
    pr_id: Optional[str] = None,
    merge_pr: Optional[bool] = True,
    target_branch: Optional[str] = None,
    job_config_index: Optional[int] = None,
    update_release: bool = True,
    release_suffix: Optional[str] = None,
    package: Optional[str] = None,
    merged_ref: Optional[str] = None,
):
    options = []
    if ref:
        options += ["--ref", ref]
    if pr_id:
        options += ["--pr-id", pr_id, f"--{'no-' if not merge_pr else ''}merge-pr"]
        if merge_pr and target_branch:
            options += ["--target-branch", target_branch]
    if job_config_index is not None:
        options += ["--job-config-index", str(job_config_index)]
    if not update_release:
        options += ["--no-update-release"]
    if release_suffix:
        options += ["--release-suffix", f"'{release_suffix}'"]
    if merged_ref:
        options += ["--merged-ref", f"'{merged_ref}'"]

    # do not create symlinks in Copr environment
    options += ["--no-create-symlinks"]

    options += [url]
    return COPR_SOURCE_SCRIPT.format(
        package=f" -p {package}" if package else "",
        options=" ".join(options),
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from enum import Enum
from typing import Optional


class ActionName(Enum):
    """
    Name of the action.

    Can be defined by user in the per-package config file and used to overwrite the default
    implementation.

    New action needs to be added here and to the table in the
    `https://github.com/packit/packit.dev/content/docs/actions.md`.
    (Some values are also used in tests:
    - tests/unit/test_config.py
    - tests/unit/test_actions.py
    - tests/unit/test_base_git.py
    - tests/integration/test_base_git.py
    """

    post_upstream_clone = "post-upstream-clone"
    pre_sync = "pre-sync"
    create_patches = "create-patches"
    prepare_files = "prepare-files"
    create_archive = "create-archive"
    get_current_version = "get-current-version"
    fix_spec = "fix-spec-file"
    changelog_entry = "changelog-entry"
    commit_message = "commit-message"

    @classmethod
    def is_valid_action(cls, action: str) -> bool:
        return action in cls.get_possible_values()

    @classmethod
    def get_action_from_name(cls, action: str) -> Optional["ActionName"]:
        return None if not cls.is_valid_action(action) else ActionName(action)

    @classmethod
    def get_possible_values(cls):
        return [a.value for a in ActionName]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import time
from os import getcwd

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import JobType, pass_config
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)

logger = logging.getLogger("packit")


@click.command("in-image-builder")
@click.option(
    "--job-config-index",
    default=None,
    type=click.INT,
    help="Use N-th job definition to load configuration for the image build. "
    "The type needs to be vm_image_build.",
)
@click.option("--wait/--no-wait", default=False, help="Wait for the build to finish")
@click.argument("image_name", default=None, type=click.STRING)
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="build"),
)
@click.argument("path_or_url", type=LocalProjectParameter(), default=getcwd())
@pass_config
@cover_packit_exception
@iterate_packages
def in_image_builder(
    config,
    wait,
    image_name,
    job_config_index,
    package_config,
    path_or_url,
):
    """
    Create a VM image in Image Builder.

    ### EXPERIMENTAL ###

    This command is experimental and the integration with Image Builder will be
    changed in a backwards incompatible way in the future.

    Packit loads image build configuration from your packit.yaml file.

    When `--job-config-index` is not specified, the job configuration is loaded from your
    .packit.yaml and the first matching vm_image_build job is used.

    IMAGE_NAME is the name of the image to be created. Please pick something unique so it's
    easy to identify for you in the Image Builder interface and can be well associated
    with the image content.

    [PATH_OR_URL] argument is a local path or a URL to the upstream git repository,
    it defaults to the current working directory
    """
    api = get_packit_api(
        config=config,
        package_config=package_config,
        local_project=path_or_url,
        job_config_index=job_config_index,
        job_type=JobType.vm_image_build,
    )

    build_id = api.submit_vm_image_build(
        image_distribution=api.package_config.image_distribution,
        image_name=image_name,
        image_request=api.package_config.image_request,
        image_customizations=api.package_config.image_customizations,
        copr_namespace=api.package_config.owner,
        copr_project=api.package_config.project,
        copr_chroot=api.package_config.copr_chroot,
    )
    logger.info(f"Image Build ID: {build_id}")
    logger.info("Browse in: https://console.redhat.com/insights/image-builder")

    if not wait:
        return 0
    logger.info(f"Getting status for build {build_id}.")
    while True:
        status = api.get_vm_image_build_status(build_id)
        if status not in ("building", "pending"):
            break
        logger.info(f"Status: {status}")
        time.sleep(5.0)

    # IB offers this link:
    # https://console.aws.amazon.com/ec2/v2/home?region=us-east-1#LaunchInstanceWizard:ami=ami-08ee0689a0547e7ea
    # we could construct & print it too



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import os
from typing import Optional

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import (
    PackageConfig,
    get_context_settings,
    pass_config,
)
from packit.config.aliases import DEPRECATED_TARGET_MAP
from packit.config.common_package_config import MockBootstrapSetup
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)
from packit.utils import sanitize_branch_name
from packit.utils.changelog_helper import ChangelogHelper

logger = logging.getLogger(__name__)


@click.command("in-copr", context_settings=get_context_settings())
@click.option("--wait/--no-wait", default=True, help="Wait for the build to finish")
@click.option(
    "--owner",
    help="Copr user, owner of the project. (defaults to username from copr config)",
)
@click.option(
    "--project",
    help="Project name to build in. Will be created if does not exist. "
    "(defaults to the first found project value in the config file or "
    "'packit-cli-{repo_name}-{branch/commit}')",
)
@click.option(
    "--targets",
    help="Comma separated list of chroots to build in. (defaults to 'fedora-rawhide-x86_64')",
    default="fedora-rawhide-x86_64",
)
@click.option(
    "--description",
    help="Description of the project to build in.",
    default=None,
)
@click.option(
    "--instructions",
    help="Installation instructions for the project to build in.",
    default=None,
)
@click.option(
    "--list-on-homepage",
    help="Created copr project will be visible on copr's home-page.",
    default=False,
    is_flag=True,
)
@click.option(
    "--preserve-project",
    help="Created copr project will not be removed after 60 days.",
    default=False,
    is_flag=True,
)
@click.option(
    "--additional-repos",
    help="URLs to additional yum repos, which can be used during build. "
    "Comma separated. "
    "This should be baseurl from .repo file. "
    "E.g.: http://copr-be.cloud.fedoraproject.org/"
    "results/rhughes/f20-gnome-3-12/fedora-$releasever-$basearch/",
    default=None,
)
@click.option(
    "--bootstrap",
    type=click.types.Choice(("default", "on", "off", "image")),
    help="mock bootstrap feature setup.",
    default=None,
)
@click.option(
    "--upstream-ref",
    default=None,
    help="Git ref of the last upstream commit in the current branch "
    "from which packit should generate patches "
    "(this option implies the repository is source-git).",
)
@click.option(
    "--request-admin-if-needed",
    help="Ask for admin permissions when we need to change settings of the copr project "
    "and are not allowed to do so.",
    default=False,
    is_flag=True,
)
@click.option(
    "--enable-net/--disable-net",
    help="Copr build is built with explicitly enabled network access or disabled",
    default=False,
    is_flag=True,
)
@click.option(
    "--release-suffix",
    default=None,
    type=click.STRING,
    help="Specifies release suffix. Allows to override default generated:"
    "{current_time}.{sanitized_current_branch}{git_desc_suffix}",
)
@click.option(
    "--default-release-suffix",
    default=False,
    is_flag=True,
    help=(
        "Allows to use default, packit-generated, release suffix when some "
        "release_suffix is specified in the configuration."
    ),
)
@click.option(
    "--module-hotfixes",
    help="Created copr project will have module_hotfixes set to True",
    default=False,
    is_flag=True,
)
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="build"),
)
@click.argument("path_or_url", type=LocalProjectParameter(), default=os.path.curdir)
@pass_config
@cover_packit_exception
@iterate_packages
def copr(
    config,
    wait,
    owner,
    project,
    targets,
    description,
    instructions,
    list_on_homepage,
    preserve_project,
    upstream_ref,
    additional_repos,
    bootstrap,
    request_admin_if_needed,
    enable_net,
    release_suffix,
    default_release_suffix,
    package_config,
    path_or_url,
    module_hotfixes,
):
    """
    Build selected upstream project in Copr.

    PATH_OR_URL argument is a local path or a URL to the upstream git repository,
    it defaults to the current working directory.
    """
    api = get_packit_api(
        config=config,
        package_config=package_config,
        local_project=path_or_url,
    )
    release_suffix = ChangelogHelper.resolve_release_suffix(
        api.package_config,
        release_suffix,
        default_release_suffix,
    )

    if not project:
        logger.debug("Copr project name was not passed via CLI.")

        if isinstance(api.package_config, PackageConfig):
            project = api.package_config.get_copr_build_project_value()

        if project:
            logger.debug("Using a first Copr project found in the job configuration.")
        else:
            logger.debug(
                "Copr project not found in the job configuration. "
                "Using the default one.",
            )
            sanitized_ref = sanitize_branch_name(path_or_url.ref)
            project = f"packit-cli-{path_or_url.repo_name}-{sanitized_ref}"

    logger.info(f"Using Copr project name = {project}")

    targets_list = targets.split(",")
    for target in targets_list:
        if target in DEPRECATED_TARGET_MAP:
            logger.warning(
                f"Target '{target}' is deprecated. "
                f"Please use '{DEPRECATED_TARGET_MAP[target]}' instead.",
            )
    targets_to_build = api.copr_helper.get_valid_build_targets(
        *targets_list,
        default="fedora-rawhide-x86_64",
    )

    logger.info(f"Targets to build: {targets_to_build}.")

    additional_repos_list: Optional[list[str]] = (
        additional_repos.split(",") if additional_repos else None
    )

    build_id, repo_url = api.run_copr_build(
        project=project,
        chroots=list(targets_to_build),
        owner=owner,
        description=description,
        instructions=instructions,
        upstream_ref=upstream_ref,
        list_on_homepage=list_on_homepage,
        preserve_project=preserve_project,
        additional_repos=additional_repos_list,
        bootstrap=MockBootstrapSetup(bootstrap) if bootstrap is not None else None,
        request_admin_if_needed=request_admin_if_needed,
        enable_net=enable_net,
        release_suffix=release_suffix,
        srpm_path=config.srpm_path,
        module_hotfixes=module_hotfixes,
    )
    click.echo(f"Build id: {build_id}, repo url: {repo_url}")
    if wait:
        api.watch_copr_build(build_id=build_id, timeout=60 * 60 * 2)
